import gzip
import sys
import tempfile
import time
import typing

import psutil
import vmprof

if sys.platform.startswith('linux'):
    from psutil._common import pcputimes
    from psutil._pslinux import pfullmem
elif sys.platform.startswith('darwin'):
    from psutil._common import pcputimes
    from psutil._psosx import pfullmem
else:
    pcputimes = typing.Tuple[float, ...]
    pfullmem = typing.Tuple[int, ...]

import django.conf
import django.http
import django.utils.timezone
import django.utils.deprecation

from .models import RequestProfile


class RequestProfiler:
    profile_file: typing.BinaryIO
    profile: RequestProfile
    process: psutil.Process
    cpu_real: float
    cpu_time: pcputimes
    cpu_diff: pcputimes
    mem_info: pfullmem
    mem_diff: pfullmem

    def __init__(self):
        self.profile_file = tempfile.TemporaryFile()
        self.process = psutil.Process()
        self.profile = RequestProfile()

    def enable(self, request: django.http.HttpRequest):
        self.profile.started_at = django.utils.timezone.now()
        self.profile.request_user_id = request.user.id
        self.profile.request_path = request.path

        self.cpu_real = time.monotonic()
        self.cpu_time = self.process.cpu_times()
        self.mem_info = self.process.memory_full_info()

        # the web view has no support for lines and memory views
        # native currently fails when trying to unload a library on next request
        # real_time is necessary for our use case - otherwise we'll only profile CPU time, which we don't care about

        vmprof.enable(self.profile_file.fileno(), lines=False, memory=False, native=False, real_time=True)

    def disable(self, response: django.http.HttpResponse):
        time_now = time.monotonic()
        self.cpu_real = time_now - self.cpu_real
        self.cpu_diff = self.cpu_time._make(q - p for (p, q) in zip(self.cpu_time, self.process.cpu_times()))
        self.mem_diff = self.mem_info._make(q - p for (p, q) in zip(self.mem_info, self.process.memory_full_info()))

        vmprof.disable()

        self.profile.created_at = django.utils.timezone.now()
        self.profile.response_code = response.status_code
        self.profile.time_real = self.cpu_real
        self.profile.time_user = self.cpu_time.user
        self.profile.time_sys = self.cpu_time.system
        self.profile.allocated_vm = self.mem_diff.vms

        self.profile_file.seek(0)
        self.profile.data = self.profile_file.read()
        self.profile_file.close()


class RequestProfilerMiddleware(django.utils.deprecation.MiddlewareMixin):
    def profile_request(self, request: django.http.HttpRequest):
        return request.user.is_authenticated()

    def profile_response(self, request: django.http.HttpRequest, response: django.http.HttpResponse, profile: RequestProfile):
        profile.data = gzip.compress(profile.data)
        profile.save()

    def process_request(self, request: django.http.HttpRequest):
        if self.profile_request(request):
            try:
                request._profiler = RequestProfiler()
                request._profiler.enable(request)
            except:
                request._profiler = None
                raise
        else:
            request._profiler = None

    def process_response(self, request: django.http.HttpRequest, response: django.http.HttpResponse):
        # we're not guaranteed to have ran .process_request() if there was an exception:
        profiler: typing.Optional[RequestProfiler] = getattr(request, '_profiler', None)
        if profiler:
            profiler.disable(response)
            self.profile_response(request, response, profiler.profile)
        return response



import io

from vmprof.profiler import read_profile

from django.contrib.auth.mixins import UserPassesTestMixin
from django.contrib.auth.models import User
from django.db.models import F, Func, When, Value, Case, BooleanField
from django.http import Http404
from django.http import JsonResponse
from django.views.generic import TemplateView
from django.views.generic.detail import DetailView
from django.views.generic.list import ListView

from .models import RequestProfile


class SuperuserRequiredMixin(UserPassesTestMixin):

    def test_func(self):
        user: User = self.request.user
        return user.is_active and user.is_superuser


class ProfilerEntryJSON(SuperuserRequiredMixin, DetailView):
    model = RequestProfile

    def render_to_response(self, context, **response_kwargs):
        data = self.object.data
        if (data is None):
            raise Http404()
        profile = read_profile(io.BytesIO(data))
        profile_tree = profile.get_tree()
        profile_data = {
            "VM": profile.interp,
            "profiles": profile_tree._serialize(),
            "argv": "%s %s" % (profile.interp, profile.getargv()),
            "version": 2,
        }
        return JsonResponse({"data": profile_data})


class ProfilerEntryView(SuperuserRequiredMixin, TemplateView):
    template_name = 'django_profiler_vmprof/entry.html'


class ProfilerIndexView(SuperuserRequiredMixin, ListView):
    model = RequestProfile
    template_name = 'django_profiler_vmprof/index.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['list'] = list(
            self.object_list
                .order_by('-started_at')
                .defer('data')
                .annotate(has_data=Case(When(data=None, then=Value(False)), default=True, output_field=BooleanField()))
                .values()
        )
        return context



#!/usr/bin/env python

import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-profiler-vmprof',
    version='0.2.0',
    install_requires=[
        'django>=1.9',
        'psutil>=5.2.1',
        'vmprof>=0.4.9',
    ],
    python_requires='~=3.6',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='A minimalist VMProf integration for Django.',
    long_description=README,
    url='https://github.com/bawr/django-profiler-vmprof',
    author='Bartosz Wróblewski',
    author_email='bawr@hszm.pl',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.9',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],
)
